/*              Tarea N3
Asignatura:     Programacion de Sistemas
Proyecto:       Monitor de trafico para Browser
NOMBRE:		mtb.c
SINOPSIS:	mtb <Puerto_Local> <IP_Proxy> <Puerto_Proxy>
DESCRIPCION:	Programa que funciona como servidor transparente entre el (browser) y el verdadero
	        Proxy que sale a internet. Ademas monitorea el trafico de Rx , Tx y Numero de
                conexiones
Alumnos:        Diego Gonzalez Barrientos
                Patricio Valle Vidal
*/

/*------------------------------------------------*/
/*Librerias incluidas y requeridas en el programa */
/*------------------------------------------------*/

//incluye variables globales, librerias y prototipos requeridos para que funcione el programa
#include "mtb.h" 

/*------------------------------------------------*/
/*Funcion principal del programa mtb              */
/*------------------------------------------------*/

int main(int argc,char *argv[] )
{
  int   sc, sp;	// descriptores de archivos para los socket del cliente y del proxy.
  struct sockaddr_in server;	// estructura para el servidor
  struct sockaddr_in client;	// estructura para el cliente
  struct sockaddr_in proxy;	// estructura para el proxy (que se ingresa en linea de comandos)
  fd_set readfds, readfdsCopy;	// listas (conjunto) de descriptores de archivos
  int n,i, clientlen, Ret;	// indicador de select, contador, tamao de estructura client, retorno de read.
  int ClientList[64], ProxyList[64], lastSockC=0, lastSockP=0;	// arreglos de conexiones, posiciones de ultimas conexiones
  char enter[10];	// almacenaje de entrada estandar

  if (argc==4)
  {
     if (signal(SIGALRM, sig_alrm) == SIG_ERR)  //Registrando evento de manejo de senales alarm
                            err_sys("signal(SIGALRM) error");
     if (signal(SIGINT, sig_int) == SIG_ERR)    //Registrando evento de manejo de senales ctrl+c
                            err_sys("signal(SIGINT) error");
     
     if (pipe(pfd) < 0)                         //Obteniendo pipa
     {	  	  
	 perror("pipe");
	 exit(1);
     }

     if ((pid = fork()) < 0)                    //Obteniendo un proceso hijo                     
     {  
	 perror("fork");
	 exit(1);
     }
     
     
     if (pid == 0)                             //Proceso hijo 
     {
	 dup2(pfd[0], 0);                      //Descriptor asociado a la salida estandar
	 close(pfd[1]);  /* close the write end of the pipe */
         execlp("scilab", "scilab", "-nw", 0);  //Proceso hijo muta a scilab
	 perror("exec");
	 exit(-1);
     }
     
     close(pfd[0]);             
     sd = fdopen(pfd[1], "w");  /* to use fprintf instead of just write */
    ////// inicio obtencion de informacion de maquinas (proxy y server) //////////

    /////////////// server ////////////

    server.sin_family = AF_INET;
    server.sin_addr.s_addr = htonl(INADDR_ANY); 	// se le asigna una direccion disponible al server
    server.sin_port = htons(atoi(argv[1])); 		// puerto del server en 1 argumento

    ////////////// proxy //////////////////

    proxy.sin_family = AF_INET;
    proxy.sin_addr.s_addr = inet_addr(argv[2]);	// direccion del proxy en 2 argumento
    proxy.sin_port = htons(atoi(argv[3])); 	// puerto del proxy en 3 argumento

    //////////// Creacion socket para cliente "sc" ////////////////

    sc = socket (AF_INET,SOCK_STREAM,0);
    if (sc < 0)
    {
      perror("opening stream socket");
      exit(-1);
    }
    reusePort(sc);


    /////////// Bindeado del socket "sc" //////////////

    if ( bind( sc, (struct sockaddr *)&server, sizeof(server) ) )/*
								se asigna la estructura "server" (con todos
								sus datos) al socket "sc"
								*/
    {
      close(sc);	// el socket "sc" se cierra si no se puede realizar el BIND
      perror("binding name to stream socket");
      exit(-1);
    }

    ///////////// Listening ////////////////////

    listen(sc,4);	// puede dejar a 4 en la lista de espera (cola)

    //////////////////////////////////////////

    clientlen = sizeof(client);	// tamao de la estructura "client"
    FD_ZERO(&readfdsCopy);	// limpia el conjunto "readfdsCopy"
    FD_SET(sc,&readfdsCopy);	// aade el descriptor dado "sc" al conjunto "readfdsCopy"
    FD_SET(0,&readfdsCopy);		// aade la entrada estandar al conjunto "readfdsCopy"

    /////////////// Ciclo de atencion de socket //////////////////////

    for(;;)
    {
      alarm(1);  //manejo de excepcion cada 1 segundo
      memcpy(&readfds, &readfdsCopy, sizeof(fd_set));	// copia area de memoria de "readfdsCopy" a "readfds"
      n = select(FD_SETSIZE, &readfds, (fd_set *) 0, (fd_set *) 0, NULL);/*
      									creo el select
									se esperan caract del listado "readfds"
									para lectura no se esperan listas de
									escritura ni de excepcion.
									*/
      if (n > 0)
      {
        if (FD_ISSET(sc, &readfds))	/* mira a ver si un descriptor del socket "sc"
					esta activado en la lista "readfds"*/
	{
	  //////////// Conexion del cliente (alimentador) con servidor ///////////////

	  ClientList[lastSockC++] = accept(sc, (struct sockaddr *)&client, &clientlen);/*
	  										 se escribe en arreglo
											 "Clientlist", en
											 posicion "lastSockC",
											 el nuevo descriptor de
											 archivo entregado por
											 accept, del socket "sc",
											 que esta pidiendo
											 atencion. Mientras se
											 realiza la coneccion
											 con el cliente y se
											 guarda la info del
											 cliente en la estructura
											 "client"
											 */
	  FD_SET(ClientList[lastSockC-1], &readfdsCopy);
	  ///////////// Conexion del servidor (mtb) con proxy ////////////////

	  //////////// Creacion socket para el proxy "sp" ////////////////

	  sp = socket (AF_INET,SOCK_STREAM,0);
	  if (sp < 0)
    	  {
	    perror("opening stream socket");
      	    exit(-1);
    	  }

	  reusePort(sp);

	  //////////// Conectando de mtb al proxy ///////////////////

	  ProxyList[lastSockP++] = sp;	// se asiga el socket "sp" al arreglo de conexiones con el proxy
	  FD_SET(ProxyList[lastSockP-1], &readfdsCopy);

	  if ( connect(sp, (struct sockaddr *)&proxy, sizeof(proxy)) < 0 )
	  {
	    close(sp);
	    perror("connecting stream socket");
	    exit(0);
	  }
	  Nconex++;
	}

        ////////// Ciclo de manejo de requerimentos cliente - proxy ///////////

	for (i=0; i < lastSockC; i++)
	{
	  if (FD_ISSET(ClientList[i], &readfds))	/* mira a ver si el descriptor "ClientList[i]"
							esta activado en la lista "readfds"*/

	    if (SocketConnector(ClientList[i],ProxyList[i],0) < 0)	/*
	    								entrada a funcion "SocketConnector"
									para leer requerimientos del cliente
									y enviarselos al proxy.
	    								*/
	    {
	      FD_CLR(ClientList[i], &readfdsCopy);	// borra el descriptor "ClientList[i]" de "readfdsCopy"
	      FD_CLR(ProxyList[i], &readfdsCopy);	// borra el descriptor "ProxyList[i]" de "readfdsCopy"
	      ClientList[i] = ClientList[--lastSockC];	// shift para descartar el socket que se desconect
	      ProxyList[i] = ProxyList[--lastSockP];	// de "ClientList" y "ProxyList"
	      i--;
	    }
	  if (FD_ISSET(ProxyList[i], &readfds))		/* mira a ver si el descriptor "ProxyList[i]"
							esta activado en la lista "readfds" */
	    if (SocketConnector(ProxyList[i],ClientList[i],1) < 0)	/*
	    								entrada a funcion "SocketConnector"
									para leer los datos enviados por proxy
									(requeridos por cliente) para enviarselos
									finalmete al cliente.
	    								*/
	    {
	      FD_CLR(ClientList[i], &readfdsCopy);	// borra el descriptor "ClientList[i]" de "readfdsCopy"
	      FD_CLR(ProxyList[i], &readfdsCopy);	// borra el descriptor "ProxyList[i]" de "readfdsCopy"
	      ClientList[i] = ClientList[--lastSockC];	// shift para descartar el socket que se desconect
	      ProxyList[i] = ProxyList[--lastSockP];	// de "ClientList" y "ProxyList"
	      i--;
	    }
        }
      }
    }     
  }
  else
  {
    printf("Error de sintasis:\n\n");
    printf("              Sintaxis: ./ppw <Puerto_Local> <IP_Proxy> <Puerto_Proxy>\n\n");
  }
}

/*----------------------------------------------------------*/
/*Funcion utilizada para realizar la conexion cliente-proxy */
/*----------------------------------------------------------*/

int SocketConnector(int lector, int escritor, int enable)
{
  char buf[512];
  int rc;

  for(;;)
  {
    if( (rc=read(lector, buf, sizeof(buf))) < 0)	/* lee la informacion proveniente del descriptor "lector"
    						 	y loguarda en "buf" (cliente o proxy)*/
      perror("receiving stream  message");
    if (rc > 0)
    {
      if (send(escritor, buf, rc, 0) <0 )	/* manda la informacion guardada en "buf" al
      						descriptor "escritor" (cliente o proxy)*/
        perror("sending stream message");
      if(enable==1)	// se asegura que los bytes analizados son los de trafico de bajada (de proxy a cliente)
      {
	bytesTX=bytesTX+rc;
      }
      else
		bytesRX=bytesRX+rc;
      return(1);
    }
    else
    {
      close (lector);	// cierra descriptor de archivos "lector"
      close (escritor);	// cierra descriptor de archivos "escritor"
      return(-1);
    }
  }
}

/*----------------------------------------------------------*/
/*Funcion utilizada para reusar el socket                   */
/*----------------------------------------------------------*/

void reusePort(int s)
{
  int one=1;

  if ( setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *) &one,sizeof(one)) == -1 )
  {
    printf("error in setsockopt,SO_REUSEPORT \n");
    exit(-1);
  }
}

/*-----------------------------------------------------------*/
/*Funcion utilizada para definir manejo de errores en senales*/
/*-----------------------------------------------------------*/

void  err_sys(char * msg) 
{   
     printf("%s \n", msg);
     exit(-1); 
}

/*----------------------------------------------------------*/
/*Funcion utilizada para manejar senal ctrl+c               */
/*----------------------------------------------------------*/

static void sig_int(int signo)
{
   
   fprintf(sd, "\n exit\n"); fflush(sd);
   waitpid(pid, &status, 0);
       fclose(sd);
   exit(0);
}

/*----------------------------------------------------------*/
/*Funcion utilizada para manejar senal alarm                */
/*----------------------------------------------------------*/

static void sig_alrm(int signo)
{
   int i;
   Ventana[(time_counter % WINDOW_SIZE)].NumConex = Nconex;
   Ventana[(time_counter % WINDOW_SIZE)].Tx = bytesTX;
   Ventana[(time_counter % WINDOW_SIZE)].Rx = bytesRX;   
   printf("Tiempo transcurrido: %d[s] \n", (time_counter+1)); 
   if (time_counter == (WINDOW_SIZE - 1) || time_counter + 1  == alarm_counter)
   {
      sd = fdopen(pfd[1], "w");  
      for(i=(alarm_counter - WINDOW_SIZE);i < alarm_counter; i++)
	{
	if (((i + WINDOW_SIZE) % alarm_counter + 1) == 1)
           fprintf(sd, "y=[%d %d %d %d];\n",((i + WINDOW_SIZE) % alarm_counter + 1), i, Ventana[(i %WINDOW_SIZE)].NumConex ,  Ventana[(i % WINDOW_SIZE)].Tx,  Ventana[(i % WINDOW_SIZE)].Rx);	   
       fprintf(sd, "y(%d,:)=[%d %d %d %d];\n",((i + WINDOW_SIZE) % alarm_counter + 1), i, Ventana[(i % WINDOW_SIZE)].NumConex,  Ventana[(i % WINDOW_SIZE)].Tx,  Ventana[(i % WINDOW_SIZE)].Rx);
       }
      fflush(sd); 
      if (inside != 0)
	fprintf(sd, "xbasc();\n");      
      fprintf(sd, "subplot(3,1,1);\n");
      fprintf(sd, "plot2d(y(:,1),y(:,2),style=13);\n");
      fprintf(sd, "xtitle('Monitoreo de Conexiones','Tiempo [s]','Numero de Conexiones');\n");
      fprintf(sd, "subplot(3,1,2);\n");
      fprintf(sd, "plot2d(y(:,1),y(:,3),style=2);\n"); fflush(sd);
      fprintf(sd, "xtitle('','Tiempo [s]','Datos Transmitidos [bytes]');\n");
      fprintf(sd, "subplot(3,1,3);\n");
      fprintf(sd, "plot2d(y(:,1),y(:,4),style=5);\n"); fflush(sd);
      fprintf(sd, "xtitle('','Tiempo [s]','Datos Recibidos [bytes]');\n");fflush(sd);    
      alarm_counter = alarm_counter + alarm_refresh;
      inside = 1;      
   }
     time_counter++;
     bytesTX = 0;
     bytesRX = 0;
     Nconex = 0;
     return;
}
